#!/usr/bin/perl

######################################################################
package Cognition::HTMLParser::Utils;
######################################################################

use CGI::Util;
use Cognition::MagicString;
use Exporter;
use Roman;
use Text::Wrap;
use XML::LibXML qw(:all);
use strict;

our $BNodes    = {};

BEGIN
{
	our $Timer     = undef;
	our @ISA    = qw(Exporter);
	our @EXPORT = qw(searchClass searchRel searchID searchCXP XML_STRINGIFY STRINGIFY);
}

sub searchClass
{
	my $target = shift;
	my $dom    = shift;
	my $prefix = shift || undef;
	
	return unless ($dom);
	
	my @matches = ();
	foreach my $node ($dom->getElementsByTagName('*'))
	{
		my $classList;
		$classList = $node->getAttribute('class');
		$classList = $node->getAttribute('name')
			if ((!length $classList) && ($node->tagName eq 'param'));
			
		next unless ((length $classList));
		
		if ((defined $prefix) && $classList =~ / (^|\s) ($prefix \-?)? $target (\s|$) /x)
		{
			push @matches, $node;
		}
		elsif ($classList =~ / (^|\s) $target (\s|$) /x)
		{
			push @matches, $node;
		}
	}
	
	return @matches;
	
} #/sub searchClass


sub searchRel
{
	my $target = shift;
	my $dom    = shift;
	
	$target =~ s/[\:\.]/\[\:\.\]/;
	
	my @matches = ();
	for my $node ($dom->getElementsByTagName('*'))
	{
		my $classList = $node->getAttribute('rel');
		next unless (length $classList);
		
		if ($classList =~ / (^|\s) $target (\s|$) /ix)
		{
			push @matches, $node;
		}
	}
	
	return @matches;
	
} #/sub searchRel


sub searchID
{
	my $target = shift;
	my $dom    = shift;
	
	$target =~ s/^\#//;
	
	my @matches = ();
	for my $node ($dom->getElementsByTagName('*'))
	{
		my $id   = $node->getAttribute('id') || next;
		
		if ($id eq $target)
		{
			return $node;
		}
	}
	
} #/sub searchID


sub searchCXP
{
	my $target = shift;
	my $dom    = shift;
	
	$target =~ s/^\#//;
	
	my @matches = ();
	for my $node ($dom->getElementsByTagName('*'))
	{
		my $id   = $node->getAttribute('_xpath') || next;
		
		if ($id eq $target)
		{
			return $node;
		}
	}
	
} #/sub searchCXP


sub XML_STRINGIFY
{
	my $dom        = shift;
	my $valueClass = shift || undef;
	my $includeMe  = shift || 0;

	# Clean up.
	my $clone = $dom->cloneNode(1);
	foreach my $kid ($clone->getElementsByTagName('*'))
	{
		$kid->removeAttribute('_xpath');
		$kid->removeAttribute('_xmllang');
		$kid->removeAttribute('microformats-content');
	}
	if ($includeMe)
	{
		$clone->removeAttribute('_xpath');
		$clone->removeAttribute('_xmllang');
		$clone->removeAttribute('microformats-content');
	}
	
	# Create string
	my $rv;
	if ($includeMe)
	{
		$rv = $clone->toString;
		$rv =~ s|xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"|xmlns="http://www.w3.org/1999/xhtml"|;
		$rv =~ s|xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"||g;
	}
	else
	{
		foreach my $kid ($clone->childNodes)
		{
			my $k .= $kid->toString;
			$k =~ s|xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"|xmlns="http://www.w3.org/1999/xhtml"|;
			$k =~ s|xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"||g;
			$rv .= $k;
		}
	}
	
	# Return string
	return (ms $rv, $dom);
}


sub STRINGIFY
{
	my $dom        = shift;
	my $valueClass = shift || undef;
	my $doABBR     = shift || (length $valueClass);
	my $str;
	
	my %opts;
	
	if (ref($valueClass) eq 'HASH')
	{
		%opts = %$valueClass;
		
		$valueClass = $opts{'excerpt-class'};
		$doABBR     = $opts{'abbr-pattern'};
	}
	
	if ($opts{'rdfa-stringify'})
	{
		return _STRINGIFY_RDFa($dom, %opts);
	}
	
	return unless ($dom);

	# ABBR pattern
	if ($doABBR)
	{
		if ($dom->nodeType==XML_ELEMENT_NODE
			&& length $dom->getAttribute('microformats-content'))
		{
			my $title = $dom->getAttribute('microformats-content');
			return (ms $title, $dom);
		}
		elsif ( ($dom->nodeType==XML_ELEMENT_NODE 
			&& $dom->tagName eq 'abbr' 
			&& length $dom->getAttribute('title'))
		||   ($dom->nodeType==XML_ELEMENT_NODE 
			&& $dom->tagName eq 'acronym' 
			&& length $dom->getAttribute('title'))
		||   ($dom->nodeType==XML_ELEMENT_NODE
			&& $dom->getAttribute('title') =~ /data\:/)
		)
		{
			my $title = $dom->getAttribute('title');
	
			if ($title =~ / [\(\[] data\: (.*) [\)\]] /x
			||  $title =~ / data\: (.*) $ /x )
				{ $title = $1; }
	
			if (length $title)
				{ return (ms $title, $dom); }
		}
		elsif ($dom->nodeType==XML_ELEMENT_NODE 
			&& $opts{'datetime'} 
			&& length $dom->getAttribute('datetime'))
		{
			return (ms $dom->getAttribute('datetime'), $dom);
		}

	}
	
	# Value excerpting.
	if (length $valueClass)
	{
		my @nodes = searchClass($valueClass, $dom);
		my @strs;
		if (@nodes)
		{
			foreach my $valueNode (@nodes)
			{
				push @strs, STRINGIFY($valueNode, {
					'excerpt-class' => undef,
					'abbr-pattern'  => $doABBR,
					'datetime'      => $opts{'datetime'},
					'no-reduce-whitespace' => 1
				});
			}
			
			# In datetime mode, be smart enough to detect when date, time and
			# timezone have been given in wrong order.
			if ($opts{'datetime'})
			{
				my $dt_things = {};
				foreach my $x (@strs)
				{
					if ($x =~ /^\s*(Z|[+-]\d{1,2}(\:?\d\d)?)\s*$/i)
						{ push @{$dt_things->{'z'}}, $1; }
					elsif ($x =~ /^\s*T?([\d\.\:]+)\s*$/i)
						{ push @{$dt_things->{'t'}}, $1; }
					elsif ($x =~ /^\s*([\d-]+)\s*$/i)
						{ push @{$dt_things->{'d'}}, $1; }
				}
				
				# If multiple date-looking things, multiple time-looking things or
				# multiple timezone-looking things have been provided, this routine
				# has probably gone all wrong.
				unless (defined $dt_things->{'d'}->[1]
				     || defined $dt_things->{'t'}->[1]
				     || defined $dt_things->{'z'}->[1])
				{
					$str = sprintf("%s %s %s",
						$dt_things->{'d'}->[0],
						$dt_things->{'t'}->[0],
						$dt_things->{'z'}->[0]);
				}
			}
			
			unless (length $str)
			{
				$str = ms (join $opts{'joiner'}, @strs), $dom;
			}
		}
	}

	my $inpre = ($dom->getAttribute('_xpath') =~ /\/pre\b/i) ? 1 : 0;
	$str = _STRINGIFY($dom, $inpre, 0)
		unless (defined $str);

	unless ($opts{'no-reduce-whitespace'})
	{
		# \x1D is used as a "soft" line break. It can be "absorbed" into an adjacent
		# "hard" line break.
		$str =~ s/\x1D+/\x1D/g;
		$str =~ s/\x1D\n/\n/gs;
		$str =~ s/\n\x1D/\n/gs;
		$str =~ s/\x1D/\n/gs;
		$str =~ s/(^\s+|\s+$)//gs;
	}
	
	return (ms $str, $dom);
}


sub _STRINGIFY_RDFa
{
	my $dom  = shift;
	
	if ($dom->nodeType == XML_TEXT_NODE)
	{
		return $dom->getData;
	}
	elsif ($dom->nodeType == XML_ELEMENT_NODE && lc($dom->tagName) eq 'img')
	{
		return ms($dom->getAttribute('alt'), $dom);
	}
	elsif ($dom->nodeType == XML_ELEMENT_NODE)
	{
		my $rv = '';
		foreach my $kid ($dom->childNodes)
			{ $rv .= _STRINGIFY_RDFa($kid); }
		return ms($rv, $dom);
	}

	return '';
}


sub _STRINGIFY
{
	my $domNode   = shift || return;
	my $inPRE     = shift || 0;
	my $indent    = shift || 0;
	my $rv = '';

	my $tag;
	if ($domNode->nodeType == XML_ELEMENT_NODE)
	{
		$tag = lc($domNode->tagName);
	}
	elsif ($domNode->nodeType == XML_COMMENT_NODE)
	{
		return (ms '');
	}
	
	# Change behaviour within <pre>.
	$inPRE++ if ($tag eq 'pre');
	
	# Text node, or equivalent.
	if (!$tag || $tag eq 'img' || $tag eq 'input' || $tag eq 'param')
	{
		$rv = $domNode->getData               if (!$tag);
		$rv = $domNode->getAttribute('alt')   if ($tag
		                                      && length $domNode->getAttribute('alt'));
		$rv = $domNode->getAttribute('value') if ($tag
		                                      && length $domNode->getAttribute('value'));

		unless ($inPRE)
		{
			$rv =~ s/[\s\r\n]+/ /gs;
		}
		
		return $rv;
	}
	
	# Breaks.
	return "\n" if ($tag eq 'br');
	return "\x1D\n====\n\n"
		if ($tag eq 'hr');
	
	# Deleted text.
	return '' if ($tag eq 'del');

	# Get stringified children.
	my (@parts, @ctags, @cdoms);
	my $extra = 0;
	if ($tag =~ /^([oud]l|blockquote)$/)
	{
		$extra += 6; # Advisory for word wrapping.
	}
	foreach my $child ($domNode->getChildNodes)
	{
		my $ctag = $child->nodeType==XML_ELEMENT_NODE ? lc($child->tagName) : undef;
		my $str  = _STRINGIFY($child, $inPRE, $indent + $extra);
		push @ctags, $ctag;
		push @parts, $str;
		push @cdoms, $child;
	}
	
	if ($tag eq 'ul' || $tag eq 'dir' || $tag eq 'menu')
	{
		$rv .= "\x1D";
		my $type = lc($domNode->getAttribute('type')) || 'disc';

		for (my $i=0; defined $parts[$i]; $i++)
		{
			next unless ($ctags[$i] eq 'li');
			
			$_ = $parts[$i];
			s/(^\x1D|\x1D$)//g;
			s/\x1D+/\x1D/g;
			s/\x1D\n/\n/gs;
			s/\n\x1D/\n/gs;
			s/\x1D/\n/gs;
			s/\n/\n    /gs;

			my $marker_type = $type;
			$marker_type = lc($cdoms[$i]->getAttribute('type'))
				if (length $cdoms[$i]->getAttribute('type'));

			my $marker = '*';
			if ($marker_type eq 'circle')    { $marker = '-'; }
			elsif ($marker_type eq 'square') { $marker = '+'; }
			
			$rv .= "  $marker $_\n";
		}
		$rv .= "\n";
	}
	
	elsif ($tag eq 'ol')
	{
		$rv .= "\x1D";
		
		my $count = 1;
		$count = $domNode->getAttribute('start')
			if (length $domNode->getAttribute('start'));
		my $type = $domNode->getAttribute('type') || '1';
		
		for (my $i=0; defined $parts[$i]; $i++)
		{
			next unless ($ctags[$i] eq 'li');
			
			$_ = $parts[$i];
			s/(^\x1D|\x1D$)//g;
			s/\x1D+/\x1D/g;
			s/\x1D\n/\n/gs;
			s/\n\x1D/\n/gs;
			s/\x1D/\n/gs;
			s/\n/\n    /gs;
			
			my $marker_value = $count;
			$marker_value = $cdoms[$i]->getAttribute('value')
				if (length $cdoms[$i]->getAttribute('value'));
			
			my $marker_type = $type;
			$marker_type = $cdoms[$i]->getAttribute('type')
				if (length $cdoms[$i]->getAttribute('type'));
				
			my $marker = sprintf('% 2d', $marker_value);
			if (uc($marker_type) eq 'A' && $marker_value > 0 && $marker_value <= 26)
				{ $marker = ' ' . chr( ord($marker_type) + $marker_value - 1 ); }
			elsif ($marker_type eq 'i' && $marker_value > 0 && $marker_value <= 3999)
				{ $marker = sprintf('% 2s', roman($marker_value)); }
			elsif ($marker_type eq 'I' && $marker_value > 0 && $marker_value <= 3999)
				{ $marker = sprintf('% 2s', Roman($marker_value)); }
				
			$rv .= sprintf("\%s. \%s\n", $marker, $_);

			$count++;
		}
		$rv .= "\n";
	}

	elsif ($tag eq 'dl')
	{
		$rv .= "\x1D";
		for (my $i=0; defined $parts[$i]; $i++)
		{
			next unless ($ctags[$i] eq 'dt' || $ctags[$i] eq 'dd');
			
			if ($ctags[$i] eq 'dt')
			{
				$rv .= $parts[$i] . ':';
				$rv =~ s/\:\s*\:$/\:/;
				$rv .= "\n";
			}
			elsif ($ctags[$i] eq 'dd')
			{
				$_ = $parts[$i];
				s/(^\x1D|\x1D$)//g;
				s/\x1D+/\x1D/g;
				s/\x1D\n/\n/gs;
				s/\n\x1D/\n/gs;
				s/\x1D/\n/gs;
				s/\n/\n    /gs;
				$rv .= sprintf("    \%s\n\n", $_);
			}
		}
	}

	elsif ($tag eq 'blockquote')
	{
		$rv .= "\x1D";
		for (my $i=0; defined $parts[$i]; $i++)
		{
			next unless ($ctags[$i]);
			
			$_ = $parts[$i];
			s/(^\x1D|\x1D$)//g;
			s/\x1D+/\x1D/g;
			s/\x1D\n/\n/gs;
			s/\n\x1D/\n/gs;
			s/\x1D/\n/gs;
			s/\n\n/\n/;
			s/\n/\n> /gs;
			$rv .= "> $_\n";
		}
		$rv =~ s/> $/\x1D/;
	}
	
	else
	{
		$rv = '';
		for (my $i=0; defined $parts[$i]; $i++)
		{
			$rv .= $parts[$i];
			
			# Hopefully this is a sensible algorithm for inserting whitespace
			# between childnodes. Needs a bit more testing though.
			
			# Don't insert whitespace if this tag or the next one is a block-level element.
			# Probably need to expand this list of block elements.
#			next if ($ctags[$i]   =~ /^(p|h[1-9]?|div|center|address|li|dd|dt|tr|caption|table)$/);
#			next if ($ctags[$i+1] =~ /^(p|h[1-9]?|div|center|address|li|dd|dt|tr|caption|table)$/);
			
			# Insert whitespace unless the string already ends in whitespace, or next
			# one begins with whitespace.
#			$rv .= ' '
#				unless ($rv =~ /\s$/ || (defined $parts[$i+1] && $parts[$i+1] =~ /^\s/));
		}
		
		if ($tag =~ /^(p|h[1-9]?|div|center|address|li|dd|dt|tr|caption|table)$/ && !$inPRE)
		{
			$rv =~ s/^[\t ]//s;
			local($Text::Wrap::columns);
			$Text::Wrap::columns = 78 - $indent;
			$rv = "\x1D".Text::Wrap::wrap('','',$rv);
			if ($tag =~ /^(p|h[1-9]?|address)$/)
			{
				$rv .= "\n\n";
			}
		}
		
		if ($tag eq 'sub')
			{ $rv = "($rv)"; }
		elsif ($tag eq 'sup')
			{ $rv = "[$rv]"; }
		elsif ($tag eq 'q')
			{ $rv = "\"$rv\""; }
		elsif ($tag eq 'th' || $tag eq 'td')
			{ $rv = "$rv\t"; }
	}

	return $rv;
}


sub xmlns
{
	my $ns  = shift;
	my $dom = shift;
	
	while ($dom->nodeType==XML_ELEMENT_NODE)
	{
		if (length $dom->getAttribute("xmlns:$ns"))
		{
			return $dom->getAttribute("xmlns:$ns");
		}
		$dom = $dom->getParentNode;
	}
	
	if ($dom->nodeType==XML_DOCUMENT_NODE)
	{
		if (length $dom->getDocumentElement->getAttribute("xmlns:$ns"))
		{
			return $dom->getDocumentElement->getAttribute("xmlns:$ns");
		}
	}
	
	return undef;
}

sub PrepareXPointers
# We save the node paths into special @_xpath attributes. Why? Because we often
# need to move DOM nodes about (e.g. uF include pattern) but still want access
# to the original node paths. Sometimes we're just dealing with document
# fragments, which have no meaningful node paths. This seems like an inefficient
# way of doing it, but I can't find a better solution.
{
	my $dom = shift;
	foreach my $tag ($dom->getElementsByTagName('*'))
		{ $tag->setAttribute('_xpath', $tag->nodePath); }
}

sub PrepareXmlLang
{
	my $dom  = shift;
	my $lang = shift || '';
	my $http = shift || undef;
	
	return unless ($dom->nodeType==XML_ELEMENT_NODE);
	
	if (length $dom->getAttribute('lang'))
	{
		$lang = $dom->getAttribute('lang');
	}
	elsif (length $dom->getAttribute('xml:lang'))
	{
		$lang = $dom->getAttribute('xml:lang');
	}
	elsif (lc($dom->tagName) eq 'html')
	{
		my @metas = $dom->getElementsByTagName('meta');
		foreach my $m (@metas)
		{
			next
				unless ($m->getAttribute('http-equiv') =~ /^\s*(content\-language)\s*$/i);
			$lang = $m->getAttribute('content');
			last;
		}
		$lang = $http->header('content-language')
			if ($http && !$lang);
		$lang = $1 if ($lang =~ /\s*([^\,\;]+)/);
	}
		
	$dom->setAttribute('_xmllang', $lang);
	
	foreach my $kid ($dom->getChildrenByTagName('*'))
	{
		PrepareXmlLang($kid, $lang);
	}
}

sub XPointer
{
	my $dom  = shift;
	my $page = shift || '';
	
	my $xpath = $dom->getAttribute('_xpath');
	if (!length $xpath)
		{ $xpath = $dom->nodePath; }
	
	my $escapedXPath = CGI::Util::escape($xpath);
	$escapedXPath =~ s|\%2f|/|ig;
	$escapedXPath =~ s|\%3f|?|ig;
	$escapedXPath =~ s|\%5b|[|ig;
	$escapedXPath =~ s|\%5d|]|ig;
	return "$page#xpointer($escapedXPath)";
}

sub BNode
{
	my $dom    = shift || undef;
	my $prefix = shift || 'Node';
	my $suffix = shift || 1;
	my $fmt    = '%s%02d';
	
	if ($dom)
	{
		my $xpath;
		if (ref $dom)
			{ $xpath = $dom->getAttribute('_xpath'); }
		else
			{ $xpath = $dom; }
		
		foreach my $key (keys %$Cognition::HTMLParser::Utils::BNodes)
		{
			if ($Cognition::HTMLParser::Utils::BNodes->{$key} eq $xpath
			&&  $key =~ /^ $prefix [0-9\.]+ $/x)
			{
				return $key;
			}
		}

		my $i = $suffix;
		while (defined $Cognition::HTMLParser::Utils::BNodes->{sprintf($fmt, $prefix, $i)}) { $i++; }
		$Cognition::HTMLParser::Utils::BNodes->{ sprintf($fmt, $prefix, $i) } = $xpath;
		return sprintf($fmt, $prefix, $i);
	}
	
	my $i = 1;
	while (defined $Cognition::HTMLParser::Utils::BNodes->{sprintf($fmt, $prefix, $i)}) { $i++; }
	$Cognition::HTMLParser::Utils::BNodes->{sprintf($fmt, $prefix, $i)} = 'anon';
	return sprintf($fmt, $prefix, $i);
}

sub BNodeURI
{
	return 'bnode:///' . BNode(@_);
}


1; #/package Cognition::HTMLParser::Utils




